include(as86.m4)

; BIOS-based Loader for the 32-bit MOOSE API.
; Copyright (c) Franois-Ren "Far" Bn Rideau Dang-Vu 1994,1995

; Please Read the info/BIOS-API.nfo file (section * loader.asm::).


include(API.m4)			; API header structure


;;; Debugging defines:

__NO_USE_INT_13H__=0
DEBUG=1				; debug messages
__DEBUG__=1			; heavy debug
CLEARSCREEN=1			; clear it before loading
MESSAGES=1			; Do print messages



;;; Defines that may change.

; We mustn't really use these. The Boot Record need it, though.
LOADER_SEG_SIZE	 =  0x0F00	;(60KB) ; size in para for the loader segment.
LOADER_STACK_TOP =  LOADER_SEG_SIZE*16	; SP in the same segment
STACKSIZE	 =  8192		; stack size when using LILO
LOADER_SECTS	 =  3			; size of the loader in sectors.
LOADER_OFFSET	 =  0x100		; load the loader as a .COM file.
LOADER_UDATA	 =  0			; no udata area, as 80h-100h is used.
KERNEL_SECTS	 =  512			; Size of Kernel+API to load.
MIN_KERNEL_SPACE =  300			; KB needed by the kernel
MAX_KERNEL_SPACE =  16000		; KB asked by the kernel
MAX_TRY		 =  10		        ; number of retries before aborting I/O

;;; Defines that won't change
LOG_SECT_SIZE	=	9		; Size of sectors (low-level standard)
SECT_SIZE	=   1 << LOG_SECT_SIZE	; 512
MAP_FILE	 =  LOADER_OFFSET+SECT_SIZE*LOADER_SECTS+LOADER_UDATA
			; just after the loader

; LOAD MODEs:
; * 'D' for loading a DOS file as the API+Kernel. LoadArg points to the
;  name (11 bytes) of the system file containing the API and the Kernel
;  system means the sectors are contiguous and unmovable, as with usual
;  DOS files...
; * 'H' for using an overformatted floppy's hidden sectors as the API+Kernel.
;  LoadArg contains real max tracks (hi) and real max sects/tracks (lo).
; * 'C' for using contiguous directories after the current one. LoadArg is
;  The number of the first sector (relative to hidden_sectors).
; * 'M' for LILO-like mapping. The map file is already loaded at address
;  MAP_FILE
; * 'L' for just moving a LILO (or LOADLIN) loaded kernel. Why rewrite LILO ?


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Code first ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
USE16


		.text
		.org	LOADER_OFFSET	; .COM file format

.globl _main
_main:
		;jmp	near Init	; Init just follows.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Loader DATA ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
.data

;.bss	;;I manually allocated data; I should have used macros...

newheap(0,0x80)
ConvMem		=	allot(0,4)	; Total Conventional memory in bytes
AvConvMem	=	allot(0,4)	; Available Conv. mem in bytes
XtdMem		=	allot(0,4)	; Total Extended memory in bytes
AvXtdMem	=	allot(0,4)	; Available Extended mem in bytes

BaseConvMem:	.long	0x1000  ;4096	; Page aligned, reserve 2K for BIOS.
BaseXtdMem	.long	0x110000;1MB+64KB ; Page aligned, reserve low
					; memory & HMA for dos emulator.

;;; Inits data for Next_cluster
Init_Sector_Read =	allot(0,2)

;;; Returns the next block of sector to read in int 13h -able registers.
;;; al=0 if no more clusters.
Next_cluster	=	allot(0,2)


;;; we assume we could fill the user-defined buffer
;;; if not, then bufferize those routines with an overfillable
;;; buffer (overfilling: up to the maximal value given by Next_cluster).
over_fill	=	allot(0,4)
over_fill_ofs	=	over_fill
over_fill_seg	=	over_fill+2
over_fill_len	=	allot(0,2)

;;; We also assume no more than 64 KB overfilled buffer, no 64 KB boundary
;;; crossed, no pointer wraps, new value of buffer not interfering with old
;;; overfull buffer. (that's all :)


;;; Load Mode 'M'
pCluster	=	allot(0,2)


;;; Load Mode 'H' ; for fdformatted floppies only
;;; (can anyone achieve over-formatted hd's ?)
CurrentTrack	=	allot(0,2)	; hi: track, lo: head
Max_Tracks_on	=	allot(0,1)	; max. number of "official" tracks

;;; Load Mode 'C'
			heapalign(0,4)
CurrentSector	=	allot(0,4)
MaxSector	=	allot(0,4)

;;; Debugging LOADER
Int13Handler	=	allot(0,4)

;;; Load Mode 'D'
DataArea	=	allot(0,2)
CurrentCluster	=	allot(0,2)
SectorInCluster	=	allot(0,2)
FAT_Reader	=	allot(0,2)	; routine to read 12 or 16 bit FAT
FAT_Mapped_Beg	=	allot(0,2)	; for 16 bit FATs, which part is mapped
FAT_Mapped_Len	=	allot(0,2)


;;; Load Mode 'L'
; Nothing more is needed ;

CPUType 	=	allot(0,2)	; AH:CPU, AL:NPU (see mem386.inc)
ErrorCode	=	allot(0,1)	; must be all on the same disk.



;;; API Header information	; see APIHeader.asm

API_Signature	=	0x1000	; Signature_Struc  es:0x1000
API_Header	=	0x1020	; API_Header_Struc es:0x1020
API_Magic	=	API_Signature+AS_Magic_Number	; dword
API_Length	=	API_Header+AH_TotalLength	; dword
API_EntryPoint	=	API_Header+AH_EntryPoint	; dword

;;; LILO style loading
INITSEG		=	0x9000
SYSSEG		=	0x1000
SETUPSEG	=	0x9020
SYSSIZE		=	0x7F00
Lxsig_1		=	0xAA55
Lxsig_2		=	0x5A5A

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ARGUMENTS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; see bootrec.asm for full details...

;;; DOS info table

bytes_per_sect	=	0x0B	; word
sectors_per_clu	=	0x0D	; byte
reserved_sect	=	0x0E	; word
num_FATs	=	0x10	; byte
entries_rootdir	=	0x11	; word
total_num_sect	=	0x13	; word
media_descr	=	0x15	; byte
sect_per_FAT	=	0x16	; word
sect_per_track	=	0x18	; word
num_heads	=	0x1A	; word
hidden_sect	=	0x1C	; dword
d_total_n_sect	=	0x20	; dword

;;; MOOSE Loading Parameters

LoadMode	=	0x24	; byte
LoadArg		=	0x25	; word
BIOS_DL		=	0x27	; byte
Clusters	=	0x28	; dword

;;; Kernel Header

Kernel_Magic	=	0x00	; dword es:0h
Kernel_Length	=	0x20	; dword es:20h


;;;;;;;;;;;;;;;;;;;;;;;;; Initialization Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
.text

;;;;; These procedures are executed only once, so I macro'ed them.

define({InitConfused},
{
	STOP(fs,0,38,14,'C',0x0A)
	call I_Confd
   I_Confd:
	STOP(fs,0,39,14,'o',0x0A)
	mov bx,cs
	pop dx	; return address
	mov cx,#I_Confd
	sub cx,dx
	sar cx,#4
	sub bx,cx
	mov ds,bx
	z_z
	STOP(fs,0,40,14,'n',0x0A)
	lsa(si,"I dunno where I am !",13,10,0)
	STOP(fs,0,41,14,'f',0x0A)
	z_z
Exit2:
	jmp Exit
})

define({LILO_test},
{
	STOP(fs,0,39,13,'O',0x0F)
	__IF__(nz)
	  call L_t
	 AfterJump:
         Init_Confused:
	   InitConfused
         L_t:
	  pop ax			; Get Return Address
	  cmp ax,#AfterJump
	__THEN__
	  mov bx,cs
	  cmp bx,#SETUPSEG
	  jnz Init_Confused
	  cmp ax,#AfterJump-0x100
	  jnz Init_Confused
	  .byte 0xea		; far jump
	  .word LILOused,SETUPSEG-0x10
        LILOused:
	;;; Init registers for LILO
 	  STOP(fs,0,37,13,'L',0x09)
	  mov ax,cs
	  mov ds,ax
	  mov es,ax
	  mov ss,ax
	  mov sp,#BufferBeg+STACKSIZE
	  z_z					; strings not loaded yet.
 	  STOP(fs,0,37,14,'I',0x09)
	;;; Load what remains of the loader (as under linux)
	  End_LILOs_work
	  z_z({"Readjusted from LILO:",13,10},SST) ; strings not loaded yet.
	;;; Init parameters for LILO
  	  STOP(fs,0,37,15,'L',0x9)
	  mov [over_fill_seg],bx
	  mov [LoadMode],#'L'
	  mov [byte 80h],#0x99
	  ;; one day, add support for Linuxish command-line arguments...
 	  STOP(fs,0,37,16,'O',0x9)
	__ENDIF__
})

define({End_LILOs_work},{	; stolen from linux/boot/setup.S
; POSTcondition:
; * the Loader is fully placed in memory
; * bx contains the beginning of the system loaded by LILO
; Check signature at end of setup
	STOP(fs,0,1,5,'S',0x9)
	push	ds
	mov	bx,#SYSSEG
	cmp	[setup_sig1],#Lxsig_1
	jne	bad_sig
	cmp	[setup_sig2],#Lxsig_2
	jne	bad_sig
	jmps	good_sig

; We now have to find the rest of the setup code/data
bad_sig:
	STOP(fs,0,2,5,'B',0x09)
	mov	ax,#INITSEG
	mov	es,ax
	xor	dh,dh
	seg es
	mov	dl,[497]	; get setup sects from boot sector
	sub	dx,#4		; LILO loads 4 sectors of setup
	shl	dx,#8		; convert to words
	mov	cx,dx
	shr	dx,#3		; convert to segment
	add	bx,dx

; Move rest of setup code/data to here
	STOP(fs,0,3,5,'M',0x09)
	mov	di,2048		; four sectors loaded by LILO
	clr	si
	mov	ax,#SETUPSEG
	mov	es,ax
	mov	ax,#SYSSEG
	mov	ds,ax
	;z_z
	rep
	movsw

; restore ds
	STOP(fs,0,4,5,'R',0x09)

; that's all, folks
good_sig:
	STOP(fs,0,5,5,'G',0x09)
	pop	ds
	STOP(fs,0,6,5,'D',0x09)
})


define({Init_LOADER},{
;;; Test CPU & FPU type
  inlinelib(TestCPU)
  mov [CPUType],ax
  __IF__(b)
    cmp ah,#3
  __THEN__
    lsa(si,{'MOOSE currently needs a 32 bit (386+) processor'})
    jmp near Exit
  __ENDIF__
 IF __NO_USE_INT_13H__
    push es
    push word 0
    pop es
    seg es
    mov eax,[0x13*4]
    mov [Int13Handler],eax
    pop es
})


define({Parse_Options},{
;;; Verify that DOS aint loaded
   STOP(fs,0,40,13,'O',0xF)
   __IF__(z)
	cmp [byte 0x80],#0x99	; The bootrec ought to have put this value
				; 99h >=80h, thus COMMAND.COM won't put it. '
   __THEN__
      lsa(si,{"From DOS, use LOADLIN !",CRLF})
      jmp Puts
   __ENDIF__
   mov al,[LoadMode]
   z_z
   mov bx,#D_Init_disk
   mov dx,#D_Next_cluster
   cmp al,#'D'
   jz Parsed
   mov bx,#M_Init_disk
   mov dx,#M_Next_cluster
   cmp al,'M'
   jz Parsed
   mov bx,#C_Init_disk
   mov dx,#C_Next_cluster
   cmp al,'C'
   jz Parsed
   mov bx,#H_Init_disk
   mov dx,#H_Next_cluster
   cmp al,'H'
   jz Parsed
   mov bx,#L_Init_disk
   mov dx,#L_Next_cluster
   cmp al,'L'
   jz Parsed
 Load_Mode_Unsupported:
   __dataseg__
     Mode_not_supported:	.ascii "Load Mode "
     Unsupported_Mode:		.ascii "X not supported."
     db_zero:			.byte 0
   __codeseg__
   mov [Unsupported_Mode],al
   mov si,#Mode_not_supported
   jmp near Exit
 Parsed:
   mov [Init_Sector_Read],bx
   mov [Next_cluster],dx
})

define({Init_Message},{
	__dataseg__
	  sLoading:	.ascii "Loading using Mode "
	  Loading_Mode:	defb('X',"..",0)
	__codeseg__
	mov al,[LoadMode]
	mov [Loading_Mode],al
	mov si,#sLoading
	PUTS
})

define({Get_Memory},{
   STOP(fs,0,41,13,'S',0xF)
  ;;; Get Total extended memory
	mov ah,#0x88
	int #0x15	; result in KB
	movzx eax,ax
	shl eax,#10
	mov [XtdMem],eax; result in bytes
  ;;; Compute available x. mem.
	sub eax,#65536	; Don't put anything in the HMA zone '
	mov [AvXtdMem],eax
  ;;; Get Total conventional memory
	int #0x12	; result in KB
	movzx eax,ax
	shl eax,#10
	mov [ConvMem],eax
IF 0
  ;;; Compute Available conventional memory
    ;- assuming the boot record loaded us as high as possible) is *not*
    ; possible, since LILO could have done otherwise
	shr eax,#4
	sub ax,#0x1000	; 64KB
	push ds
	mov ds,ax
	mov ebx,#0xFFFC
	cmp [dword bx],#MAGIC_MARKER
	jnz AvConvMemFound
	dec bx
	dec bx
     Run_Res_ConvMem:
	mov cx,[bx]
	jcxz AvConvMemFound
	sub bx,cx
	jmp Run_Res_ConvMem
     AvConvMemFound:
	shl eax,#4
	add eax,ebx
	pop ds
	mov [AvConvMem],eax
ENDIF
})

define({Init_Disk_Read},{
   ;;; Init sector count
	call [Init_Sector_Read]
})

define({LOAD_API},{
   ; Note: for the moment, the API should not be larger than 60 KB,
   ;  and even 30 KB to avoid a 64KB DMA limit with sector prefetching ;
   ;  Else a BIOS wrapper will have to buffer sector reading somewhere
   ;  (the 2KB between 00800h and 01000h ?) when some wrapping is detected.

   STOP(fs,0,42,13,'E',0xF)
   ;;; Init The buffer
	clr ax
	mov es,ax	; 4KB
	mov bx,#0x1000
   ;;; Load the Header
	mov al,#64	; Header length ((;) ah=0 already)
	call DoReadOnce
   ;;; Get the length from the header
	__IF__(z)
	  cmp [dword es:bx],#MAGIC_PC_API ; API_Magic
	__THEN__
	  zdwoz("Magic number ",[es:bx],{CRLF})
	  z_z
	  lsa(si,{"API file fails magic number test.",CRLF})
	  jmps Exit3
	__ENDIF__
	mov eax,[API_Length]
	__IF__(a)
	  cmp eax,0xF000 ; we cannot it if bigger than 60K.
	__THEN__
	  lsa(si,"API Too big",CRLF)
	Exit3:
	 z__
	  jmp Exit
	 ELSE
	  jmp Exit
	 __z
	__ENDIF__

   ;;; Load the entire API.
	add bx,#64
	sub ax,#64
	call DoReadOnce
   ;;; ... ToDo:
   ;;; - add relocation (?)/ header parsing
   ;;; - add decompression (?)
})

Macro	LUKI_GO_API
   ;;; Jump into the API.
	PUTS <".",CRLF>
	clr bx
	push bx
	mov es,bx
	mov ds,bx
	push [word API_EntryPoint]
	; (;) we should test that it is inferior to the API size ;
	IF DEBUG
	  z_z
	  push ds
	  pusha
	  push cs
	  pop ds
	  @Zone@ <"API Entry:">,es,<[word API_EntryPoint]>,16,<>
	  GETC <CRLF,"Hit any key to go MOOSE !",CRLF>
	  popa
	  pop ds
	ENDIF
	retf
EndM	LUKI_GO_API

Macro	Do_Exit
	call PrCRLF
	PUTS
	PUTS	<CRLF,"Hit any key to reboot">
	GETC
EndM	Do_Exit

Proc	Init	near
P8086

IF __DEBUG__
	P386
	push 0B800h
	pop fs
	STOP fs:,38,13,'M',0Fh
ENDIF

IF CLEARSCREEN
	mov	ax,0003h	; 80x25, 16 colors
	int	10h
ENDIF

 ;__ <"Testing if using Linux-ish loading",CRLF> ; Won't work if using LILO ;
	LILO_test

IF DEBUG
  P186	; for __ 's pusha/popa
ENDIF
 __ <"Testing z_z",CRLF>
	z_z

 __ <"Initing self",CRLF>
	Init_LOADER
	@wo@ <"CPU:">,[CPUType],<CRLF>

P386

IF 0
 __ <"Testing Int 13h",CRLF>
	push 0
	pop es
	@addr@ <" at ">,<[word es:13h*4+2]>,<[word es:13h*4]>,<CRLF>
	mov ah,0
	mov dl,0
	Int13h
	z_z
ENDIF

 __ <"Looking at options",CRLF>
	Parse_Options

 __ <"Hello, World !",CRLF>
	Init_Message

 __ <"Seeing how much memory is left",CRLF>
	Get_Memory
        @dwo@ <"Found 0x">,[ConvMem],<" bytes of conventional memory,",CRLF>
        @dwo@ <"and 0x">,[XtdMem],<" bytes of extended memory",CRLF>

 __ <"Initialize disk reading",CRLF>
	Init_Disk_Read

 __ <"Load the API at 4KB",CRLF>
	LOAD_API
	@dwo@ <"Done loading a 0x">,[API_Length],<" long API",CRLF>

 __ <"Jumping into the API",CRLF>
	LUKI_GO_API

Exit:
 __ <"Aborting...",CRLF>
	Do_Exit			; On error, exit
	ret
EndP	Init

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Console I/O ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
codeseg

IFE MESSAGES

Putc	EQU	Return
Puts	EQU	Return
PrCRLF	EQU	Return
cGetc	EQU	Return
Return:	ret

ELSE

Putc	EQU	cPutc
Getc	EQU	cGetc

; Adapted from MicroSoft's own boot sector
; DS:SI=ASCIIZ string to print
; AX spoilt, SI points just after the string-ending zero, 7 put into BX.
Proc	Puts	near
	;mov	al,[cs:si]
	;inc	si
	cld
	lodsb
	or	al,al
	jz	Return
	PUTC
	jmps	Puts
EndP	Puts

IF 0
Proc	Beep
; Echo a BEEP to the the Console
	mov	al,7
;	jmps	cPutc
EndP	Beep
ENDIF

; Outputs a character to console
; spoils AX and DI, keeps si
Proc	cPutc
	Pushm	bx,si,di,es	; no pusha: may be used on 8088 ;;;
	mov	ah,0Eh	; function Teletype
	mov	bx,7	; attribute
	int	10h	; Video BIOS call
	Popm	bx,si,di,es
Return:
	retn
EndP	cPutc

IF DEBUG
;%%OUT Watch out: z__
;z__

dataseg
  RegsMsg	db "DI   SI   BP   SP   "
		db "BX   DX   CX   AX   "
		db "FLGS CS   DS   ES   "
		db "SS   IP   STK1 STK2"
		db CRLF,0
  ERegsMsg1	db "EDI      ESI      EBP      ESP      "
		db "EBX      EDX      ECX      EAX      EFLAGS",CRLF,0
  ERegsMsg2	db "CS   DS   ES   SS   IP   STK1 STK2 STK3 "
		db "STK4 STK5 STK6 STK7 STK8 STK9 STKA STKB"
		db CRLF,0
codeseg

Proc	DumpERegs
	push ss
	push es
	push ds
	push cs
	pushfd
	pushad
	push ss
	pop es
	push cs
	pop ds
	mov si,#ERegsMsg1
	PUTS
	mov di,sp
	mov cx,9
	call PutLZone
	mov si,#ERegsMsg2
	PUTS
	mov cx,16
	call PutWZone
	popad
	popfd
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
 Return2:
	ret
EndP	DumpERegs

Proc	DumpRegs
	push ss
	push es
	push ds
	push cs
	pushf
	pusha
	push ss
	pop es
	push cs
	pop ds
	mov si,#RegsMsg
	PUTS
	mov di,sp
	mov cx,16
	call PutWZone
	call PrCRLF
	popa
	popf
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
	ret
EndP	DumpRegs

Proc	PutLZone_
	call	Space
PutLZone:
	mov eax,[es:di]
	add di,4
	call PutHexDword
	loop PutLZone_
	ret
EndP	PutLZone_

Proc	PutWZone_
	call	Space
PutWZone:
	mov ax,[es:di]
	inc di
	inc di
	call	PutHexWord
	loop PutWZone_
	ret
EndP	PutWZone_
Proc	PutZone_
	call	Space
PutZone:
	mov al,[es:di]
	inc di
	call	PutHexByte
	jcxz	short Return ;2
	call	Space
	loop PutZone_
	ret
EndP	PutZone_

Proc	PutAddr
;; Prints the contents of es:di
	push	di
	mov	ax,es
	call	PutHexWord
	PUTC	':'
	pop	ax
	jmps	PutHexWord
EndP	PutAddr

Proc	PutHexDword
	rol	eax,16
	call	PutHexWord
	rol	eax,16
;	jmps	PutHexWord
EndP	PutHexDword

PutHexWord:
; Prints the contents of AX in hexa.
; everything but F saved
		xchg	al,ah
		call	PutHexByte
		xchg	al,ah
;		jmp	PutHexByte

ENDIF
;__z
;%%OUT __z reached

; Prints the contents of AL in hexa.
; saves AX.
Proc	PutHexByte
	push	cx
	mov	cl,4
	rol	al,cl
	call	PutHexNybble
	rol	al,cl
	pop	cx
;	jmp	PutHexNybble
EndP	PutHexByte

; Prints to BIOS console the hexa nybble in low AL
; saves AX, puts 7 into BX
Proc	PutHexNybble
	push	ax
	and	al,0Fh
	add	al,'0'
	cmp	al,0Ah+'0'
	jb	do_putnyb
	add	al,'A'-'0'-0Ah
 do_putnyb:	
	PUTC
	pop	ax
	retn
EndP	PutHexNybble

Proc	cGetc
	clr	ah
	int	16h
	retn
EndP	cGetc

IF 0
Proc	sPutc
	stosb
	retn
EndP	sPutc
Proc	sGetc
	lodsb
	retn
EnP	sGetc
ENDIF

Proc	PrCRLF	near
	PUTC	13
	mov	al,10
	jmp	Putc
EndP	PrCRLF

IF DEBUG ;z__
Proc	Space
	mov al,' '
	jmp Putc
EndP	Space
ENDIF;__z

ENDIF	; MESSAGES

;;;;;;;;;;;;;;;;;;;;;;;;;; Misc. Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

codeseg

Macro	_z_ m,s
   IF DEBUG
	push ds
	push ss
	pop ds
	pusha
	PUTS <m>
	popa
	pop ds
	z_z <>,s
   ENDIF
EndM	_z_

Proc	safe_movsb
	jcxz	Return_0
safe_movsb_:
; compare source & destination
	_z_ <"safe_movsb:",CRLF>
	push esi
	push edi
	push eax
	push ebx
	clr eax
	clr ebx
	mov ax,ds
	mov bx,es
	shl eax,4
	shl ebx,4
	movzx esi,si
	movzx edi,di
	add esi,eax
	add edi,ebx
	cmp edi,esi
	pop ebx
	pop eax
	pop edi
	pop esi
	ja sm_dest_gt_source
;sm_source_gt_dest:
	_z_ <"source gt dest:",CRLF>		; I couldn't write "s>d" ):
	cld
	push	cx
	__IF__(z)
	  shr	cx,2
	__THEN__
	  rep	movsd
	__ENDIF__
	pop	cx
	__IF__(z)
	and	cx,3
	__THEN__
	  rep	movsb
	__ENDIF__
Return_0:
	ret
; assuming dest > source
 sm_dest_gt_source:
	_z_ <"source < dest:",CRLF>
	add	si,cx
	add	di,cx
	sub	si,4
	sub	di,4
	std
	push	cx
	__IF__(z)
	  shr	cx,2
	__THEN__
	  rep	movsd
	__ENDIF__
	pop	cx
	add	si,3
	add	di,3
	__IF__(z)
	  and	cx,3
	__THEN__
	  rep	movsb
	__ENDIF__
	ret
EndP	safe_movsb

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; DISK PROCESSING ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
IF __NO_USE_INT_13H__
  Macro Int13h
    __ <"Int 13h called:",CRLF>
    z_z
    pushf
    push cs
    call Do_Int13h
    __ <"Int 13h returned:",CRLF>
    z_z
  EndM
  Proc Do_Int13h
    cli
    jmp [Int13Handler]
  EndP
ELSE
  Macro Int13h
    __ <"Int 13h called:",CRLF>
    z_z
    Int 13h
    __ <"Int 13h returned:",CRLF>
    z_z
  EndM
ENDIF


;;;;;;;;;;;;;;;;;;;;;;;;;;;; General Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc	SectorNumToBIOS_Args
; In: eax (0 for first sector)
; Out: BIOS sector identifier (cl,dh,ch)

	;__ <"SectorNumToBIOS_Args Input:",CRLF>
	;z__z

	clr edx
	movzx ebx,[byte sect_per_track]
	div ebx
	mov cl,dl	; remaining: sector
	inc cl		; they are from one on
	movzx ebx,[byte num_heads]
	clr edx
	div ebx
	mov dh,dl	; remaining: head
	mov ch,al
	shl ah,6
	or cl,ah

	;__ <"SectorNumToBIOS_Args Output:",CRLF>
	;z__z
	ret
EndP	SectorNumToBIOS_Args

;Proc	BIOSToSectorNum
; In: BIOS sector identifier
; Out: eax
;	dec cl			; sectors begin with 1
;	movzx ebx,cl		; ebx := sector number
;	and bl,3Fh
;	shr cl,6		; ax := cylinder number
;	mov al,ch
;	mov ah,cl
;	movzx ecx,dh		; ecx := head number
;	mul [byte num_heads]	; dx,ax := cylinder number * heads
;	push dx			; eax := dx,ax
;	push ax
;	pop eax
;	add eax,ecx
;	movzx edx,[byte sect_per_track]
;	mul edx
;	add eax,ebx
;	ret
;EndP	BIOSToSectorNum

Proc	Read_Sectors_With_Retries
;; ax,bx,cd,dx registers unchanged.
;; si (?), di spoilt
;; CF set if error (then see [ErrorCode])
 ;__ <"Reading Sectors",CRLF>
 ;z_z
   mov di,MAX_TRY	; maximum number of retries
   mov ah,02h	; Read
  Read_Try:
   push di
   push ax
   Int13h
   mov [ErrorCode],ah
   pop ax
   jnc Read_ok
   push ax
   clr ah		; reset controller
   Int13h
   pop ax
   pop di
   dec di
   jnz Read_Try
 ;Too_Many_Retries:
   stc
   ;@by@ <"Disk I/O Error 0x">,[ErrorCode],<CRLF>
   ret
 Read_ok:
   pop di
   ret
EndP	Read_Sectors_With_Retries

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Proc	Read_Once
; Input:
; * (Min) Length to read in ax
; * buffer in es:bx
; Output:
; * buffer (overwritten)
; * if buffer changed place, data moved, assuming no interference
; * ax: preserved
; * es:bx unchanged
; * cx: data actually read (overfull if > ax; end of file if < ax)
; * si,di,dx spoilt ; bp, ds preserved
; *** NOTE: if you use data over bx+ax, you must set [over_fill_len] to
; ***  zero, or set it to another good value, and modify [over_fill_ofs]
; * other registers spoilt

 ;;; That's Structured programming in assembly ;
 __ <"Read Once Input: ",CRLF>
 @wo@ <"asked 0x">,ax,<" bytes at ">
 @addr@ <>,es,bx,<CRLF>

 ;;; load old buffer address
   mov si,[over_fill_ofs]
   mov dx,[over_fill_seg]
   mov cx,[over_fill_len]
 @wo@ <"Already were 0x">,cx,<" bytes at ">
 @wo@ <>,dx,<":">
 @wo@ <>,si,<CRLF>

   jcxz RO_Buffer_ok
 ;;; compare to new one
   __IF__(t)
     mov di,es
     cmp di,dx
    ;@OR nz ; not supported yet
     jnz RO_new_buf_dif
     cmp si,bx
   __THEN__	; Then move old to new
    RO_new_buf_dif:
     push ds
     push cx
     mov ds,dx
     mov di,bx
     z_z
     call safe_movsb ; es:bx,ax unchanged
     pop cx
     pop ds
   __ENDIF__

 RO_Buffer_ok:
 __ <"Buffer now ok",CRLF>
   push bx
   @WHILE b	; While buffer strictly not overfull enough
     cmp cx,ax
   @DO		; Do fill it again
     push ax
     push cx
     push bx		; bx: buffer address
     call [Next_cluster]; al,cx,dx: next cluster to read.
     pop bx
     or al,al		; al: number of sectors to read
     jz End_Of_File
     push ax
     call Read_Sectors_With_Retries
     pop dx		; now, it's dl
     pop cx
     pop ax
     jc Disk_IO_Error
     __ <"Cluster Read",CRLF>
     z_z
     shl dx,LOG_SECT_SIZE ; (;) wrong if LOG_ < 8
     add cx,dx
     add bx,dx
   @WEND
 ;; full enough, then give the caller what he asked
   pop bx ; original buffer
   ;pop ax ; original length asked for
   mov dx,cx ; length actually read.
   sub dx,ax
   mov [over_fill_len],dx
   @wo@ <"Overfill length now 0x">,dx,<>
   mov dx,bx
   add dx,ax
   mov [over_fill_ofs],dx
   mov [over_fill_seg],es
   @addr@ <" at ">,es,dx,<CRLF>
   __ <"Done Reading once",CRLF>
   z_z
   ret

 End_Of_File:
   __ <"End of file reached",CRLF>
   z_z
   pop cx ; total unsufficient length of data read
   pop ax ; length asked
   pop bx ; old buffer value
   mov dx,bx
   add dx,cx
   mov [over_fill_seg],es
   mov [over_fill_ofs],dx
   mov [over_fill_len],0
   ret

;;; On Errors,
 Disk_IO_Error:
   @PrByte <CRLF,"Error 0x">,[ErrorCode],<CRLF>
   mov si,#db_zero
   pop bx ; original bx
   pop ax ; return address to Main
   jmp Exit
EndP	Read_Once

Proc DoReadOnce
; calls Read_Once with abortion in case of end-of-file reached.
; See Read_Once just above for specifications.
   call Read_Once
   cmp cx,ax
   jb Premature_end_of_file
   ret
 Premature_end_of_file:
   lsa si,<"Unexpected end of file",CRLF>
   pop ax; return address
   jmp near Exit
EndP DoReadOnce

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'M' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc M_Init_disk near
    mov [pCluster],MAP_FILE
    ret
EndP M_Init_disk

Proc M_Next_cluster near
	push si
	mov si,[pCluster]
	lodsw
	mov cx,ax
	lodsw
	mov dh,ah
	__IF__(z)
	  or al,al
	__THEN__
	  mov [pCluster],si
	__ENDIF__
	pop si
	mov dl,[BIOS_DL]
	ret
;Entry format:
;BIOS_CX	dw	0002h		; CH:cyl&0xFF, CL:sect+((cyl>>8)<<6)
;BIOS_AL	db	02h		; Int 13h parameters to load the LOADER
;BIOS_DH	db	00h		; head
EndP M_Next_cluster

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'D' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Proc D_Init_disk near
; Unsupported yet
	mov al,'D'
	jmp Load_Mode_Unsupported
; Look in the root directory, search for the system file, see length and
; cluster.

; ...

EndP D_Init_disk

Proc D_Next_cluster near

; memorized part of cluster number is next sectors to load.
; while we have a contiguous (without even track limit) zone of sectors on
; disk, look in the FAT for next cluster
; When we reach a limit (or when finished), load the contiguous part.
; memorize next part of cluster to load

; Look in the FAT for the cluster number
; if 12 bit FAT, all the FAT fits in buffer
; if 16 bit FAT, have bits of it; happily a single entry won't be split
; on two sectors.

; ...
EndP D_Next_cluster

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'H' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The LoadParam is used as sect/trk (low) & #cyl (hi).

Proc H_Init_disk near
; Leave track 0 head 0 for the boot sector + fake boot sector + loader
; floppies have less than 255 tracks and 2 heads, (thus) less than 65535 sects.
   mov [word CurrentTrack],0100h	; already read Cyl 0 (lo), Hd 1 (hi)
   mov ax,[total_num_sect]
   shr ax,1 ; number of heads=1 ; cf should be 0
   div [byte sect_per_track]
   ; we should have ah=0 (all sectors used in declared surfaces&cylinders)
   mov [Max_Tracks_on],al
   ret
EndP H_Init_disk

Proc H_Next_cluster near
   mov ax,[LoadArg]	; real max tracks & max sects/tracks
   mov dx,[CurrentTrack]
   __IF__(z)
     xor dh,1	; change head
   __THEN__
     inc dl	; change track
   __ENDIF__
   cmp dl,[Max_Tracks_on]
   ja  extra_track
   mov cl,[byte sect_per_track]
   sub al,cl
   inc cl
 H_done:
   mov [CurrentTrack],dx
   mov ch,dl
   mov dl,[BIOS_DL]
   ret
 extra_track:
   mov cl,1
   cmp dl,ah	; only ah real tracks.
   jbe H_done
   clr al
   ret
EndP H_Next_cluster


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'C' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
C_Init_disk:
	__({"'C' Init disk",CRLF})
	mov ebx,[hidden_sect]
  ; we store the next sector to read in CurrentSector
	movzx eax,[LoadArg]
	add eax,ebx
	mov [CurrentSector],eax
  ; we store the maximal value to read in MaxSector
	movzx eax,[reserved_sect]
	add eax,ebx
	mov [MaxSector],eax
	;z__z
	ret

C_Next_cluster:
	xor al,al
	mov esi,[CurrentSector]
	mov edi,[MaxSector]
	sub edi,esi
	jb CNcret
	push esi
	mov eax,esi
	call SectorNumToBIOS_Args
	mov dl,cl ; sector
	and dl,#0x3F
	mov al,[byte sect_per_track]
	sub al,dl
	inc al
	movzx eax,al
	__IF__(a)
	  cmp eax,edi
	__THEN__
	  mov eax,edi
	__ENDIF__
	pop esi
	add esi,eax
	mov [CurrentSector],eax
	mov dl,[BIOS_DL]

;__ <"'C' Next Cluster Output:",CRLF>
	;z__z

  CNcret:
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'L' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
L_Init_disk:
	__({"'L' Init disk",CRLF})
	;mov [over_fill_seg],#0x1000	; (;) set in End_LILOs_work
	mov [over_fill_ofs],#0
	mov [over_fill_len],#0x8000	; (let's say) 32K max API+kernel length
					; so we're safe
	ret

L_Next_cluster:
	__({"'L' Next cluster",CRLF})
	xor al,al	; end_of_file
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;; Low-Level Disk Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;

; From Linux:
; * This procedure turns off the floppy drive motor, so
; * that we enter the kernel in a known state, and
; * don't have to worry about it later.
kill_motor:
	push dx
	mov dx,#0x3f2
	xor al,al
	out dx,al
	pop dx
	ret



__end__		; undivert et all

;;;;;;;;;;;;;;;;;;;;;;;; Signature for Linuxish loading ;;;;;;;;;;;;;;;;;;;;;;
__dataseg__
  .align 4

setup_sig1	.word	Lxsig_1
setup_sig2	.word	Lxsig_2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; End of LOADER.com ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
BufferBeg:
